#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2002 Free Software Foundation
#
# FILE:
# GFForm.py
#
# DESCRIPTION:
# Class that contains the internal python object representation of a GNUE Form
# built from GFObjects
#
# NOTES:
#

import sys
import string
import traceback

from gnue.common import GDebug, GConnections, GTrigger
from gnue.common import events
from gnue.common.FileUtils import openResource
from GFObjects import *
from gnue.common.GDataObjects import ConnectionError as DBError
from gnue.common.GRootObj import GRootObj
from gnue.common.GTrigger import TriggerError
from gnue.common.GParser import GImportItem
from GFLibrary import *
import GFLibrary
import GFParser

# Defines which objects are "Tab Stops"
TabStops = ('GFEntry','GFButton')

class GFForm(GRootObj, GFObj, events.EventAware):
  def __init__(self, parent=None, app=None):
  # TODO: with a little tweaking we can now stop passing  GFParser.getXMLelements
    GRootObj.__init__(self, 'form', GFParser.getXMLelements, GFParser)
    GFObj.__init__(self, parent)
    self._type = "GFForm"
    self._currentObject = [self]

    self._databaseDictionary = {}
    self._datasourceDictionary = {}

    # Dictionary to locate named triggers
    self._triggerDictionary = {}

    # Insert/Overwrite mode
    self._insertMode = 1

    # Focus information
    self._currentPage = None
    self._currentBlock = None
    self._currentEntry = None

    self._blockList = []
    self._pageList = []

    self._app = app

    self._triggerns={}

    # The "None" init gives datasources time to setup master/detail
    self._inits = [self.primaryInit, None, self.secondaryInit]

    #
    # New trigger support
    #
    self._validTriggers = { 'ON-STARTUP':     'On-Startup',
                            'ON-EXIT':        'On-Exit',
                            'PRE-COMMIT':     'Pre-Commit',
                            'POST-COMMIT':    'Post-Commit' }

    self._triggerGlobal = 1
    self._triggerFunctions = {'setFocus':{'function':self.triggerSetFocus,
                                          'global': 1,
                                          },
                              'getAuthenticatedUser':{'function':self.getAuthenticatedUser,
                                          'global': 1,
                                          },
                              'getCurrentEntry':{'function':self.triggerGetCurrentEntry,
                                          'global': 1,
                                          },
                              'getCurrentBlock':{'function':self.triggerGetCurrentBlock,
                                          'global': 1,
                                          },
                              'getCurrentPage':{'function':self.triggerGetCurrentPage,
                                          'global': 1,
                                          },
                              'setStatusText':{'function':self.triggerSetStatusText,
                                          'global': 1,
                                          },
                              'getParameter':{'function':self.getParameter,
                                          'global': 1,
                                          },
                              'setParameter':{'function':self.setParameter,
                                          'global': 1,
                                          },
                              'showMessage':{'function':self.showMessageBox,
                                          'global': 1,
                                          },
                              'commit':{'function':self.commit,
                                          'global': 1,
                                          },
                              'close':{'function':self.close,
                                          'global': 1,
                                          },
                              'getFeature':{'function':self.triggerGetFeature,
                                          'global': 1,
                                          },
                              'setFeature':{'function':self.triggerSetFeature,
                                          'global': 1,
                                          }
                              }

    self._features = {
      'GUI:MENUBAR:SUPPRESS': 0,
      'GUI:TOOLBAR:SUPPRESS': 0
    }


  #
  # Convert some deprecated options to new style
  #
  def _buildObject(self):
    i = 0
    while i < len(self._children):
      if isinstance(self._children[i], GFOptions):
        j = 0
        while j < len(self._children[i]._children):
          if self._children[i]._children[j].name in \
             ('title','width','height'):
            self.__dict__[self._children[i]._children[j].name] = \
                 self._children[i]._children[j].value
            self._children[i]._children.pop(j)
          else:
            j = j + 1
        if not len (self._children[i]._children):
          self._children.pop(i)
        else:
          i = i + 1
      else:
        i = i + 1

    return GFObj._buildObject(self)

  #
  # primaryInit
  #
  # Called during phaseInit startup by GParser.
  #
  def primaryInit(self):

    # Initialize our events system
    events.EventAware.__init__(self, self._app.eventController)

    self.initTriggerSystem()
    self._triggerns.update(self._triggerNamespaceTree._globalNamespace)
    # TODO: Is this right
    #self._triggerns['runForm'] = self._app.manager.runFormFromTrigger

  #
  # Secondary init (stuff that must happen
  # after all our children have init'ed)
  #
  def secondaryInit(self):
    # create the first records
    for key in self._datasourceDictionary.keys():
      if not self._datasourceDictionary[key].hasMaster():
        self._datasourceDictionary[key].createEmptyResultSet()

    # Set initial focus
    self.findAndChangeFocus(self)
    self.processTrigger('On-Startup')


  #
  # Get a user parameter. If parameter not specified, pull default value
  #
  def getParameter(self, parameter):
    param = string.lower(parameter)
    try:
      return self._parameters[param]
    except KeyError:
      rv = None
      for child in self._children:
        if isinstance(child, GFParameter) and child._name == param:
          rv = child.default
          self._parameters[param] = rv
          break
    return rv

  #
  # Set a user parameter.
  #
  def setParameter(self, parameter, value):
    param = string.lower(parameter)
    self._parameters[param]=value


  #
  # show a MessageBox from a trigger
  #
  def showMessageBox(self,msgtxt):
    message = GFMsgBox(self, msgtxt)
    message.show()


  #
  # Find focus
  #
  # Find the next focusable item given
  # an item of type Page, Block, or Entry
  #
  def findFocus(self, object):
    entry = None

    if isinstance(object, GFObj):
      if (object._type in TabStops and
          (not object.hidden) and ((not object.readonly) or
         (self._currentBlock and \
          self._currentBlock.mode=='query' and \
          object._queryable))):
        return object
      else:
        for child in object._children:
          entry = self.findFocus(child)
          if entry:
            break

    return entry

  #
  # findAndChangeFocus
  #
  # Runs findFocus and then runs
  # changeFocus with that value.
  #
  def findAndChangeFocus(self, object):
    entry = self.findFocus(object)
    if entry:
      self.changeFocus(entry)


  #
  # changeFocus
  #
  # changes to the requested entry object requested by
  # an event source
  #
  def changeFocus(self, widget, fireFocusTriggers=1):

    if not hasattr(widget,'_block'):
      return None

    message = None

    try:
      if self._currentEntry:
        event = events.Event('endEDITMODE', None)
        self._app.dispatchEvent(event)
        if event.__error__:
          return 1

      fieldChange = widget != self._currentEntry
      blockChange = widget._block != self._currentBlock
      pageChange = widget._block._page != self._currentPage

      if fireFocusTriggers:
        try:
          if fieldChange:
            self._currentEntry.processTrigger('Pre-FocusOut')
          if blockChange:
            self._currentBlock.processTrigger('Pre-FocusOut')
          if pageChange:
            self._currentPage.processTrigger('Pre-FocusOut')
            self._currentPage.processTrigger('Post-FocusOut')
          if blockChange:
            self._currentBlock.processTrigger('Post-FocusOut')
          if fieldChange:
            self._currentEntry.processTrigger('Post-FocusOut')
        except AttributeError:
          pass

      oldEntry = self._currentEntry

      self._currentEntry = widget
      self._currentBlock = self._currentEntry._block
      self._currentPage = self._currentBlock._page

      if pageChange:
        self.dispatchEvent(events.Event('gotoPAGE',self._currentPage));

      if blockChange:
        self.refreshDisplay(self._currentBlock)

      self.dispatchEvent(events.Event('updateENTRY', oldEntry))
      self.dispatchEvent(events.Event('updateENTRY', self._currentEntry))

      if fireFocusTriggers:
        if pageChange:
          self._currentPage.processTrigger('Pre-FocusIn')
        if blockChange:
          self._currentBlock.processTrigger('Pre-FocusIn')
        if fieldChange:
          self._currentEntry.processTrigger('Pre-FocusIn')
          self._currentEntry.processTrigger('Post-FocusIn')
        if blockChange:
          self._currentBlock.processTrigger('Post-FocusIn')
        if pageChange:
          self._currentPage.processTrigger('Post-FocusIn')

    except TriggerError, t:
      GDebug.printMesg(1, "Trigger Error!")
      message = _("Trigger Exception :\n") + t.msg
    return message

  #
  # end of routines made for walking
  #


  def updateUIEntry(self,entry):
    self._app.dispatchEvent(events.Event('updateENTRY',entry));

  #
  # Incomming Event handlers
  #

  def newRecord(self):
    if not self.readonly:
      self._currentBlock.newRecord()

  def deleteRecord(self):
    if not self.readonly:
      self._currentBlock.deleteRecord()

  def isSaved(self):
    for block in self._blockList:
      if not block.isSaved():
        return 0
    return 1

  def prepQuery(self):
    message = None
    if self._currentBlock.mode != 'query':
      for block in self._blockList:
        if not block.isSaved():
          message = _('Data not saved. Save changes or clear the form to proceed.')
          return message

    for block in self._blockList:
      block.processRollback()
      for entry in block._entryList:
        if hasattr(entry,'queryDefault'):
          GDebug.printMesg(1, "%s will be set to %s" % (entry.name, entry.queryDefault))
          entry.setValue(entry.queryDefault)
    self._currentBlock.initQuery()
    return message

  def executeQuery(self):
    self._app.dispatchEvent(events.Event('beginWAIT',None));

    message = None
    try:
      self._currentBlock.processQuery()

    except DBError:
      self.rollback(1)
      message = _("Database query error:\n%s\n%s ") % (sys.exc_info()[0], sys.exc_info()[1])

    self._app.dispatchEvent(events.Event('endWAIT',None));

  def commit(self):
    self._app.dispatchEvent(events.Event('beginWAIT',None));

    message = None

    try:
      if not self.readonly:
        # Form level pre-commit
        try:
          self.processTrigger('Pre-Commit')
        except TriggerError:
          GDebug.printMesg(1, "Trigger form Pre-Commit threw a TriggerError!")
          self._app.dispatchEvent(events.Event('endWAIT',None));
          return _("Form trigger returned error")

        for block in self._blockList:
          GDebug.printMesg(1, "Saving %s"%block.name)
          try:

            # This gets lost in the Pre-Commit code
            block._precommitRecord = block._currentRecord

            block.processCommit()
          except TriggerError:
            GDebug.printMesg(1, "Trigger block Pre-Commit threw a TriggerError!")
            self._app.dispatchEvent(events.Event('endWAIT',None));
            return _("Block trigger returned error")
          block.processTrigger('Post-Commit')

        for block in self._blockList:
            block.switchRecord(0)

      else:
        message = _('Form is readonly')
        self._app.dispatchEvent(events.Event('endWAIT',None));
        return message

      self.processTrigger('Post-Commit')

    except DBError:
      self.rollback(1)
      message = _("Database commit error:\n%s\n%s ") % (sys.exc_info()[0], sys.exc_info()[1])
    except:
      print _("\n\nGFForm: Unexpected Exception:")
      print '-'*60
      traceback.print_exc(file=sys.stdout)
      print '-'*60

    self._app.dispatchEvent(events.Event('endWAIT',None));

    return message

  def executeAbout(self,event):
     version = VERSION
     name = self.getOption('title') or "Unknown"
     formversion = self.getOption('version') or "Unknown"
     author = self.getOption('author') or "Unknown"
     description = self.getOption('description') or "Unknown"

     self.dispatchEvent(events.Event('uiABOUT',[version,name,formversion,author,description]))

  def rollback(self, recover=0):
    self.endEditing()
    for block in self._blockList:
      block.processRollback(recover)
    self.refreshDisplay(self)
    self._currentBlock.jumpRecord(self._currentBlock._currentRecord)

  #
  # fireTrigger
  #
  def fireTrigger(self, triggerName):
    self._triggerDictionary[triggerName](self)

#
# Updated methods for the new gnue-forms code
#


  #
  # nextEntry
  #
  # Called whenever an event source has requested that the
  # focus change to the next data entry object
  #
  def nextEntry(self):
    nextEntry = None
    firstEntry = None
    keepNext = 0

    for object in self._currentEntry._block._focusOrder:
      # Put the first field as the next to rollover
      if ((not object.hidden) and ((not object.readonly) or
         (self._currentBlock.mode=='query' and object._queryable))):
        if nextEntry == None:
          nextEntry = object
          firstEntry = object

        if object == self._currentEntry:
          keepNext = 1
          continue

        if keepNext == 1:
          nextEntry = object
          break

    if nextEntry == firstEntry and self._currentBlock.transparentBlock:
      # Jump to the next block if block is tagged as transparent
      self.nextBlock()
    else:
      self.changeFocus(nextEntry)

  #
  # previousEntry
  #
  # Called whenever an event source has requested that the
  # focus change to the next data entry object
  #
  def previousEntry(self):
    # find last entry widget
    nextEntry = None
    lastEntry = self._currentEntry

    for object in self._currentBlock._focusOrder:
      if ((not object.hidden) and ((not object.readonly) or
         (self._currentBlock.mode=='query' and object._queryable))):
        nextEntry = object
        lastEntry = object

    keepNext = 0
    for object in self._currentBlock._focusOrder:
      # Put the first field as the next to rollover
      if ((not object.hidden) and ((not object.readonly) or
         (self._currentBlock.mode=='query' and object._queryable))):
        if object == self._currentEntry:
          break

        nextEntry = object

    if self._currentBlock.transparentBlock and nextEntry == lastEntry: # Jump to the prev block if block is tagged as transparent
      prevBlock = self.findPreviousBlock()
      # Move to the new last record of the new current block
      for object in prevBlock._focusOrder:
        if not object.hidden and (not object.readonly or self._currentBlock.mode=='query'):
          nextEntry = object

    self.findAndChangeFocus(nextEntry)


  #
  # refreshDisplay
  #
  # Refreshes all the UI fields to match
  # the virtual form's fields
  #
  def refreshDisplay(self, block):
    block.walk(self.__refreshDisplay)

  def __refreshDisplay(self, object):
    if object._type == 'GFEntry':
      self.dispatchEvent(events.Event('updateENTRY',object))



  #
  # nextBlock
  #
  # Called whenever an event source has requested that the
  # focus change to the next data entry block
  #
  def nextBlock(self):

    try:
      nextBlock = self._blockList[self._blockList.index(self._currentBlock)+1]
    except IndexError:
      nextBlock = self._blockList[0]

    self.findAndChangeFocus(nextBlock)


  #
  # findPreviousBlock
  #
  # Convenience routine to find the previous block
  #
  def findPreviousBlock(self):
    try:
      return self._blockList[self._blockList.index(self._currentBlock)-1]
    except IndexError:
      return self._blockList[-1]


  #
  # previousBlock
  #
  # Called whenever an event source has requested that the
  # focus change to the next data entry block
  #
  def previousBlock(self):
    self.findAndChangeFocus(self.findPreviousBlock())


  #
  # endEditing
  #
  # Signal to the current Entry to stop editing
  # mode and save it's value to the virtual form
  #
  def endEditing(self):
    event = events.Event('endEDITMODE',None)
    self.dispatchEvent(event)
    return not event.__error__


  def prevRecord(self):
    if self._currentBlock.mode == 'query':
      GDebug.printMesg(5,'Cannot go to previous record: in query mode!')
      return
    self._currentBlock.prevRecord()

  def firstRecord(self):
    if self._currentBlock.mode == 'query':
      GDebug.printMesg(5,'Cannot go to first record: in query mode!')
      return
    self._currentBlock.firstRecord()

  def lastRecord(self):
    if self._currentBlock.mode == 'query':
      GDebug.printMesg(5,'Cannot go to last record: in query mode!')
      return
    self._currentBlock.lastRecord()

  def nextRecord(self):
    if self._currentBlock.mode == 'query':
      GDebug.printMesg(5,'Cannot go to next record: in query mode!')
      return
    self._currentBlock.nextRecord()

  def jumpRecord(self,count):
    if self._currentBlock.mode == 'query':
      GDebug.printMesg(5,'Cannot jump to record: in query mode!')
      return
    self._currentBlock.jumpRecord(count)

  def toggleInsertMode(self):
    self._insertMode = not self._insertMode

  def triggerSetFocus(self,object):
    self.changeFocus(object._object,1)
    self.dispatchEvent(events.Event('gotoENTRY',{'object':self._currentEntry}))

  def triggerGetCurrentEntry(self):
    return self._currentEntry

  def triggerGetCurrentBlock(self):
    return self._currentBlock

  def triggerGetCurrentPage(self):
    return self._currentPage

  def triggerSetStatusText(self,tip=''):
    self.dispatchEvent(events.Event('uiUPDATESTATUS',[tip,None,None,None,None,None,None]))

  # Close this form.
  def close(self):
    event = events.Event('requestEXIT')
    self._app.dispatchEvent(event)
    if event.__error__:
      return 1

  def triggerSetFeature(self, feature, value):
    if not self._features.has_key(feature):
      raise KeyError, "Trigger attempted to set unknown feature %s" % feature
    else:
      self._features[feature] = value

  def triggerGetFeature(self, feature):
    try:
      return self._features[feature]
    except KeyError:
      raise KeyError, "Trigger attempted to get unknown feature %s" % feature

  def getAuthenticatedUser(self, connection=None):
    return self._app.connections.getAuthenticatedUser(connection)
