#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2001-2002 Free Software Foundation
#
# FILE:
# GFInstance.py
#
# DESCRIPTION:
# Client startup file that parses args, builds form from file
# and passes control to a UI.
#
# NOTES:
#  Once all the events are moved back in here
#    debug
#    make the next/prec methods more generic in the GFForm
#    change self._form in here to a list

import os
import sys
from gnue.forms.GFForm import *
from gnue.forms.GFParser import loadFile
from gnue.forms import VERSION
from gnue.common import GDebug, GDataObjects
from gnue.common import events

from gnue.common.GClientApp import *

class GFInstance(events.EventAware):
  #
  # Initialize the class
  #
  def __init__(self, manager, serial, connections, ui, disableSplash=0,
               parameters={}):

    self.eventController = events.EventController()

    events.EventAware.__init__(self, self.eventController)

    self.connections = connections
    self.manager = manager
    self._serial = serial
    self._uimodule = ui
    self._disableSplash = disableSplash
    self._parameters = parameters

    # Incoming events
    self.registerEventListeners( {
                           # Datasource trigger events
                           'preCommit'           : self.preCommit,
                           'preUpdate'           : self.preUpdate,
                           'preInsert'           : self.preInsert,
                           'preDelete'           : self.preDelete,
                           'onRecordTouched'     : self.onRecordTouched,

                           # Focus-related events
                           'requestNEXTENTRY'    : self.nextEntry,
                           'requestPREVENTRY'    : self.previousEntry,
                           'requestNEXTBLOCK'    : self.nextBlock,
                           'requestPREVBLOCK'    : self.previousBlock,
                           'requestFIRSTRECORD'  : self.firstRecord,
                           'requestLASTRECORD'   : self.lastRecord,
                           'requestPREVRECORD'   : self.prevRecord,
                           'requestNEXTRECORD'   : self.nextRecord,
                           'requestRECORDNUMBER' : self.jumpToRecord,
                           'requestJUMPRECORD'   : self.jumpRecords,
                           'requestPAGE'         : self.gotoPage,
                           'requestFOCUS'        : self.changeFocus,

                           # Data set manipulation
                           'requestNEWRECORD'    : self.newRecord,
                           'requestMARKFORDELETE': self.deleteRecord,
                           'requestCOMMIT'       : self.executeCommit,
                           'requestPRINTOUT'     : self.executePrintout,
                           'requestROLLBACK'     : self.executeRollback,
                           'requestENTERQUERY'   : self.requestQuery,
                           'requestEXECQUERY'    : self.executeQuery,

                           # Miscellaneous stuff
                           'requestEXIT'         : self.executeExit,
                           'requestABOUT'        : self.executeAbout,
                           'requestMODETOGGLE'   : self.toggleInsertMode,
                           'fireTRIGGER'         : self.fireTrigger,

                           # "Entry" events (proxied)
                           'requestKEYPRESS'     : self.proxyEntryEvent,
                           'requestREPLACEVALUE' : self.proxyEntryEvent,
                           'requestCURSORMOVE'   : self.proxyEntryEvent,
                           'requestCURSORLEFT'   : self.proxyEntryEvent,
                           'requestCURSORRIGHT'  : self.proxyEntryEvent,
                           'requestCURSOREND'    : self.proxyEntryEvent,
                           'requestCURSORHOME'   : self.proxyEntryEvent,
                           'requestBACKSPACE'    : self.proxyEntryEvent,
                           'requestDELETE'       : self.proxyEntryEvent,
                           'requestTOGGLECHKBOX' : self.proxyEntryEvent,
                           'beginEDITMODE'       : self.proxyEntryEvent,
                           'endEDITMODE'         : self.proxyEntryEvent,

                           # Selection/clipboard events (proxied)
                           'requestSELECTALL'    : self.proxyEntryEvent,
                           'requestSELECTTOHOME' : self.proxyEntryEvent,
                           'requestSELECTTOEND'  : self.proxyEntryEvent,
                           'requestSELECTLEFT'   : self.proxyEntryEvent,
                           'requestSELECTRIGHT'  : self.proxyEntryEvent,
                           'requestCOPY'         : self.proxyEntryEvent,
                           'requestCUT'          : self.proxyEntryEvent,
                           'requestPASTE'        : self.proxyEntryEvent,

                     } )

  #
  # Associate a form object with this instance
  #
  def setForm(self, form):
    self._form = form

  #
  # Return this instance's serial number
  #
  def getSerialNumber(self):
    return self._serial

  #
  # Return this instance's manager
  #
  def getManager(self):
    return self.manager

  #
  # proxyEntryEvent
  #
  # This serves as a proxy for any entry-level events.
  # Since we don't want all entries listening for all
  # events (only the *current* entry), we will capture
  # any relevant events here and pass on to the entry.
  #
  def proxyEntryEvent(self, event):

    if self._form._currentEntry != None and \
       self._form._currentEntry._type != 'GFButton':

      handler = self._form._currentEntry._displayHandler

      # If the display will need to be refreshed,
      # then the proxied event should set this to 1
      # TODO: Maybe this needs set to 0 and the handlers adjusted - jst
      event.refreshDisplay = 1

      # Pass off the event to the current entry's event handler
      handler.subevents.dispatchEvent(event)

      # Refresh entry display if appropriate
      if event.refreshDisplay:
        handler.generateRefreshEvent()
        self.updateRecordStatus()

      # If the entry needs an error message displayed,
      # then the proxied event should set this to the
      # message text
      if event.__errortext__:
        messageBox = GFMsgBox(self, event.__errortext__)
        messageBox.show()


  #
  # nextEntry
  #
  # Called whenever an event source has requested that the
  # focus change to the next data entry object
  #
  def nextEntry(self, event):
    if not self._form.endEditing():
      return
    message = self._form.nextEntry()
    if message:
      messageBox = GFMsgBox(self, message)
      messageBox.show()
      return
    self.dispatchEvent(events.Event('gotoENTRY',{'object':self._form._currentEntry}))
    self.updateRecordStatus()
    self.updateTip()

  #
  # prevEntry
  #
  # Called whenever an event source has requested that the
  # focus change to the next data entry object
  #
  def previousEntry(self, event):
    if not self._form.endEditing():
      return
    self._form.previousEntry()
    self.dispatchEvent(events.Event('gotoENTRY',{'object':self._form._currentEntry}))
    self.updateRecordStatus()
    self.updateTip()

  #
  # nextBlock
  #
  # Called whenever an event source has requested that the
  # focus change to the next data entry block
  #
  def nextBlock(self,event):
    if not self._form.endEditing():
      return
    self._form.nextBlock()
    self.dispatchEvent(events.Event('gotoENTRY',{'object':self._form._currentEntry}))

    self.updateStatus()

  #
  # previousBlock
  #
  # Called whenever an event source has requested that the
  # focus change to the next data entry block
  #
  def previousBlock(self, event):
    if not self._form.endEditing():
      return
    self._form.previousBlock()
    self.dispatchEvent(events.Event('gotoENTRY',{'object':self._form._currentEntry}))
    self.updateStatus()

  #
  # gotoPage
  #
  # Called to make the form jump to a specific page
  #
  def gotoPage(self,event):
    if not self._form.endEditing():
      return
    self._form.findAndChangeFocus(self._form._pageList[event.data])
    self.dispatchEvent(events.Event('gotoENTRY',{'object':self._form._currentEntry}))
    self.updateStatus()

  #
  # _entryUpdated
  #
  # Common code snipped called when something has changed with an
  # entry and it has to update the stat bar.  Used to cut down on
  # repeated code.
  #
  def _entryUpdated(self):
    self.dispatchEvent(events.Event('gotoENTRY',{'object':self._form._currentEntry}))
    self.updateRecordCounter()
    self.updateRecordStatus()
  
  #
  # previousRecord
  #
  # Called enever an event source has requested that the
  # form advance to the next record in memory
  #
  def prevRecord(self, event):
    if not self._form.endEditing():
      return
    message = self._form.prevRecord()
    if message:
      messageBox = GFMsgBox(self,message)
      messageBox.show()
      return
    self._entryUpdated()

  #
  # nextRecord
  #
  # Called whenever an event source has requested that the
  # form advance to the next record in memory
  #
  def nextRecord(self, event):
    if not self._form.endEditing():
      return
    message = self._form.nextRecord()
    if message:
      messageBox = GFMsgBox(self,message)
      messageBox.show()
      return
    self._entryUpdated()

  #
  # firstRecord
  #
  # Called enever an event source has requested that the
  # form advance to the first record in memory
  #
  def firstRecord(self, event):
    if not self._form.endEditing():
      return
    message = self._form.firstRecord()
    if message:
      messageBox = GFMsgBox(self,message)
      messageBox.show()
      return
    self._entryUpdated()

  #
  # lastRecord
  #
  # Called enever an event source has requested that the
  # form advance to the last record in memory
  #
  def lastRecord(self, event):
    if not self._form.endEditing():
      return
    message = self._form.lastRecord()
    if message:
      messageBox = GFMsgBox(self,message)
      messageBox.show()
      return
    self._entryUpdated()

  #
  # jumpToRecord
  #
  # Called whenever an event source has requested that the
  # form move a specific record
  #
  def jumpToRecord(self,event):
    if not self._form.endEditing():
      return
    try:
      count = abs(int(event.data))-1
    except ValueError:
      message = _("Invalid numeric value entered.")
    else:
      message = self._form.jumpRecord(count)

    if message:
      messageBox = GFMsgBox(self,message)
      messageBox.show()
      return

    self._entryUpdated()

  def jumpRecords(self,event):
    if not self._form.endEditing():
      return

    count = event.data
    # Doing this to reduce comparisons in the loop
    if count > 0:
      method = self._form.nextRecord
    else:
      method = self._form.prevRecord
    for jumps in range(abs(int(count))):
      method()

    self._entryUpdated()

  def executeExit(self, event):
    if not self._form.isSaved():
      message = GFMsgBox(self, _('Data not saved. Save changes or clear the form to proceed.'))
      message.show()
    else:
      self._form.processTrigger('On-Exit')
      self.dispatchEvent(events.Event('exitApplication',_('Current data is saved')))

    return

  #
  # deleteRecord
  #
  # Tells the form to mark a record for delete
  #
  def deleteRecord(self, event):
    if self._form.readonly:
      self.dispatchEvent(events.Event('formALERT',_('Form is readonly')))
    elif self._form._currentBlock.restrictDelete:
      self.dispatchEvent(events.Event('formALERT',_('Block does not allow delete')));
    else:
      self._form.deleteRecord()
      self.updateRecordStatus()

  #
  # newRecord
  #
  # Tells the form to create a new record
  #
  def newRecord(self, event):
    if not self._form.endEditing():
      return
    if self._form.readonly:
      self.dispatchEvent(events.Event('formALERT',_('Form is readonly')))
    elif self._form._currentBlock.restrictInsert:
      self.dispatchEvent(events.Event('formALERT',_('Block does not allow insert')));
    else:
      self._form.newRecord()
      self._entryUpdated()

  #
  # executeAbout
  #
  # Displays the about dialog
  #
  def executeAbout(self,event):
    global VERSION
    version = VERSION;
    name = self._form.title or "Unknown"
    formversion = self._form.getOption('version') or "Unknown"
    author = self._form.getOption('author') or "Unknown"
    description = self._form.getOption('description') or "Unknown"
    self.dispatchEvent(events.Event('uiABOUT',[version,name,formversion,author,description]))

  #
  # toggleInsertMode(self):
  #
  # tells the form to toggle mode
  #
  def toggleInsertMode(self, event):
    self._form.toggleInsertMode()
    self.updateInsertMode()

  #
  # executePrintout
  #
  # If the form has a trigger named process-printout then fire it
  #
  def executePrintout(self,event):
    if self._form._triggerDictionary.has_key('process-printout'):
      self._form.fireTrigger('process-printout')
    else:
      # TODO: should probably do something else as a default if trigger not available
      # TODO: like a screen print
      pass

  #
  # executeRollback
  #
  # Tells the form to rollback everything it contains
  #
  def executeRollback(self,event):
    self._form.rollback()
    self._entryUpdated()

  #
  # Routines to update parts of the UI's status bar
  #
  def updateInsertMode(self):
    self.dispatchEvent(events.Event('uiUPDATESTATUS',[None,None,self._form._insertMode,None,None,None,None]))

  def updateRecordCounter(self):
    self.dispatchEvent(events.Event('uiUPDATESTATUS',[None,None,None,self._form._currentBlock._currentRecord+1,self._form._currentBlock._recordCount,None,None]))

  def updatePageCounter(self):
    maxPages = len(self._form._pageList)
    count = self._form._pageList.index(self._form._currentPage)
    self.dispatchEvent(events.Event('uiUPDATESTATUS',[None,None,None,None,None,count+1,maxPages]))

  def updateTip(self):
    tip = ''
    if self._form._currentEntry:
      if self._form._currentEntry.getOption('tip'):
        tip = self._form._currentEntry.getOption('tip')
      else:
        tip = ''

    self.dispatchEvent(events.Event('uiUPDATESTATUS',[tip,None,None,None,None,None,None]))

  def updateRecordStatus(self):
    if self._form._currentBlock.mode == 'query':
      status = 'query'
    elif self._form._currentBlock._resultSet.current.isDeleted():
      status = 'deleted'
    elif self._form._currentBlock._resultSet.current.isPending():
      status = 'modified'
    else:
      status = 'saved'

    self.dispatchEvent(events.Event('uiUPDATESTATUS',[None,status,None,None,None,None,None]))

  def updateStatus(self):
    self.updateTip()
    self.updateInsertMode()
    self.updateRecordCounter()
    self.updateRecordStatus()
    self.updatePageCounter()

  #
  # changeFocus
  #
  # changes to the requested entry object requested by
  # an event source
  #
  def changeFocus(self, event):
    if not self._form.endEditing():
      return
    if event.data._type=='GFEntry':
      if event.data.hidden:
        # is it possible ever to get into hidden field?
        return
      if (event.data.readonly and self._form._currentBlock.mode!='query'):
        GDebug.printMesg(2,"You can focus readonly field, but you won't be able to change it.")

      newEntry = event.data
      message = self._form.changeFocus(newEntry)
      if message:
        messageBox = GFMsgBox(self, message)
        messageBox.show()
        return

      self.dispatchEvent(events.Event('gotoENTRY',{'object':self._form._currentEntry}))
      self.updateRecordStatus()
      self.updateTip()

  def requestQuery(self, event):
    if not self._form.endEditing():
      return
    
#    if hasattr(self._form._currentBlock._dataSourceLink,"master"):
#      messageBox = GFMsgBox(self, _("Query from detail blocks currently not supported"))
#      messageBox.show()
#      return

    message = self._form.prepQuery()

    if message:
      messageBox = GFMsgBox(self,message)
      messageBox.show()
      return

    self._form.refreshDisplay(self._form)
    self.dispatchEvent(events.Event('gotoENTRY',{'object':self._form._currentEntry}))
    self.updateStatus()

  def executeCommit(self, event):
    if not self._form.endEditing():
      return
    message = self._form.commit()
    if message:
      messageBox = GFMsgBox(self,message)
      messageBox.show()
      return
    self._form.refreshDisplay(self._form._currentBlock)
    self._entryUpdated()

  def executeQuery(self, event):
    if not self._form.endEditing():
      return
    message = self._form.executeQuery()
    if message:
      messageBox = GFMsgBox(self,message)
      messageBox.show()
    self._form.refreshDisplay(self._form._currentBlock)
    self._entryUpdated()



  def _handlePreCommit(self, event, trigger):
    try:
      block = event.record._parent._block
    except AttributeError:
      return

    saveMode = block.mode
    block._preCommitWorkingRecord = event.record
    block.mode = 'precommit'
    block.processTrigger(trigger)
    for child in block._children:
      if child._type == 'GFEntry':
        child.processTrigger(trigger)
    block.mode = saveMode


  def preCommit(self, event):
    self._handlePreCommit(event, 'PRE-COMMIT')


  def preInsert(self, event):
    self._handlePreCommit(event, 'PRE-INSERT')


  def preUpdate(self, event):
    self._handlePreCommit(event, 'PRE-UPDATE')


  def preDelete(self, event):
    self._handlePreCommit(event, 'PRE-DELETE')


  def onRecordTouched(self, event):
    self._handlePreCommit(event, 'PRE-MODIFY')



  #
  # fireTrigger
  #
  def fireTrigger(self, event):
    self._form.fireTrigger(event.data)

  #
  #  Activate this puppy...
  #
  def activate(self):
    ui = self._uiinstance =  self._uimodule.GFUserInterface(self.eventController, self._disableSplash)
    ui.buildForm(self._form)
    ui.activateForm(self._form)

    currentEntry = self._form._currentEntry
    if not currentEntry:
      raise  _('There are no navigable widgets in this form.  Unable to display.')

    self.dispatchEvent(events.Event('gotoENTRY',{'object':currentEntry}))
    self.dispatchEvent(events.Event('updateENTRY',currentEntry))
    self._form.refreshDisplay(self._form)
    self.updateStatus()
    self._ui = ui

    # pass control to UI
    ui.mainLoop()
