#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2002 Free Software Foundation
#
# FILE:
# GFEntry.py
#
# DESCRIPTION:
# The primary data entry widget in forms
#
# NOTES:
#

from gnue.common import GDebug, FormatMasks
from gnue.forms import GFDisplayHandler
from GFValue import GFValue
from GFTabStop import GFTabStop
import string

############################################################
# GFEntry
#
# Matches the Entry form tag in the gfd
#
# It send events to its parent GFBlock
#
class GFEntry(GFValue, GFTabStop):
  def __init__(self, parent=None, value=None):
    GFValue.__init__(self, parent, value, 'GFEntry')

    # Default attributes (these may be replaced by parser)
    self.height = float(gConfig('widgetHeight'))
    self.width = float(gConfig('widgetWidth'))
    self.typecast="text"
    self.case="mixed"
    self.style="default"

    # Runtime variables
    self._uppercase = 0
    self._lowercase = 0
    self._numeric = 0
    self._queryable = 1

    self._inits = [self.initialize]
    self._allowedValues = None
    self._allowedValuesDescr = [""]

    self._rows = 1
    self._gap = 0

    #
    # Trigger exposure
    #
    self._validTriggers ={ 'PRE-FOCUSOUT':   'Pre-FocusOut',
                           'POST-FOCUSOUT':  'Post-FocusOut',
                           'PRE-FOCUSIN':    'Pre-FocusIn',
                           'POST-FOCUSIN':   'Post-FocusIn',
                           'ON-SWITCH':      'On-Switch',
                           'PRE-COMMIT':     'Pre-Commit',
                           'POST-COMMIT':    'Post-Commit',
                           'POST-QUERY':     'Post-Query',
                           'PRE-MODIFY':     'Pre-Modify',
                           'PRE-INSERT':     'Pre-Insert',
                           'PRE-DELETE':     'Pre-Delete',
                           'PRE-UPDATE':     'Pre-Update',
                           'PRE-CHANGE':     'Pre-Change',
                           'POST-CHANGE':    'Post-Change',
                         }

    self._triggerFunctions={'allowedValues':{'function':self.allowedValues,},
                            'autofillBySequence':{'function':self.triggerAutofillBySequence},
                            'isEmpty':{'function':self.isEmpty},
                            'resetForeignKey':{'function':self.resetForeignKey},
                            'set':{'function':self.triggerSetValue},
                            'get':{'function':self.getValue}
                            }
    self._triggerProperties={'readonly':{'set':self.setReadonly,
                                         'get':self.getReadonly
                                         },
                             'parent':  {'get':self.getParent}
                             }
    self._triggerSet = self.triggerSetValue
    self._triggerGet = self.getValue


  def triggerAutofillBySequence(self,sequenceName):
    if (not self.getValue()) or self.getValue()=="":
      sequenceNumber = self._block._dataSourceLink._dataObject.triggerExtensions.getSequence(sequenceName)
      self.setValue(sequenceNumber)

  def isEmpty(self):
    return self.getValue() in ("",None)

  def getReadonly(self):
    return self.readonly

  def setReadonly(self,value):
    self.readonly = value

  def _buildObject(self):
    # Convert deprecated attributes

    # Numeric attributes (now typecast="..")
    # DEPRECATED with 0.1.x
    if hasattr(self,'numeric') and self.numeric:
      del self.numeric
      self.typecast = 'number'

    # visibleCount attributes... (now rows="..")
    # DEPRECATED with 0.1.x
    if not hasattr(self, 'rows') and hasattr(self,'visibleCount'):
      self.rows = self.visibleCount
      del self.visibleCount

    # Case attributes... (now case="..")
    # DEPRECATED with 0.1.x
    if hasattr(self,'uppercase') and self.uppercase:
      del self.uppercase
      self.case = 'upper'
    elif hasattr(self,'lowercase') and self.lowercase:
      del self.lowercase
      self.case = 'lower'

    # Foreign keys
    # DEPRECATED with 0.4.x
    if hasattr(self,'foreign_key') and self.foreign_key:
      (self.fk_source, self.fk_key) = string.split(self.foreign_key,'.')
      del self.foreign_key
    if hasattr(self,'foreign_key_description') and self.foreign_key_description:
      self.fk_description = self.foreign_key_description
      del self.foreign_key_description


    # Force 'label' styles into "readonly" mode
    if self.style == 'label':
      self.readonly = 1
      self._queryable = 0

    return GFValue._buildObject(self)

  def initialize(self):
    if self.typecast == 'number':
      self._numeric = 1
      self._allowFormulas = gConfig("AllowNumericFormulas")
    if self.case == 'upper':
      self._uppercase = 1
    elif self.case == 'lower':
      self._lowercase = 1

    self._formatmask = ""
    self._inputmask = hasattr(self,'inputmask') and self.inputmask or ""
    self._displaymask = hasattr(self,'displaymask') and self.displaymask or ""

    self._block = self.findParentOfType('GFBlock')

    if hasattr(self, 'default') and self.default != None and len(self.default):
      default = self.default
    else:
      default = None

    if not hasattr(self,'field') or not len(self.field):
      self.field = "__GNUe__%s" % self.name
      self._block._dataSourceLink.referenceUnboundField(self.field, default)
      self._bound = 0
    else:
      self._block._dataSourceLink.referenceField(self.field, default)
      self._bound = 1

    # Initialize the foreign key lookup (if necessary)
    if hasattr(self,'fk_source'):
      try:
        self.fk_key
      except NameError:
        raise "fk_source specified, but no fk_key"

      try:
        self._fk_descr = string.split(self.fk_description,',')
      except NameError:
        self.fk_description = self.fk_key
        self._fk_descr = string.split(self.fk_key,',')

      self._fk_datasource = \
         self._block._form._datasourceDictionary[string.lower(self.fk_source)]

      # Reference the foreign keys to their datasources
      # (so they are selected)
      for field in [self.fk_key] + self._fk_descr:
        self._fk_datasource.referenceField(field, None)

    if hasattr(self, 'queryDefault') and \
         self.queryDefault != None and \
         self._bound and \
         len(self.queryDefault):
      self._block._queryDefaults[self] = self.queryDefault


    # Create an appropriate display handler
    if self.style == 'checkbox':
      self._displayHandler = GFDisplayHandler.CheckboxDisplayHandler(self,
               self._block._form._app.eventController)
    elif self.style == 'dropdown':
      self._displayHandler = GFDisplayHandler.DropdownDisplayHandler(self,
               self._block._form._app.eventController)
    elif self.typecast == 'text':
      self._displayHandler = GFDisplayHandler.TextDisplayHandler(self,
               self._block._form._app.eventController,
               self._displaymask, self._inputmask)
    elif self.typecast == 'number':
      self._displayHandler = GFDisplayHandler.NumberDisplayHandler(self,
               self._block._form._app.eventController,
               self._displaymask, self._inputmask)
    elif self.typecast == 'date':
      self._displayHandler = GFDisplayHandler.DateDisplayHandler(self,
               self._block._form._app.eventController,
               self._displaymask, self._inputmask)


  def isQueryable(self):
    return self._queryable


  #
  # Pulls the proper value from the resultSet
  #
  def getValue(self):
    # TODO: No longer true...
    # We have to check to see if the _block has been
    # setup yet.  The Trigger system calls getValue
    # before it's ready so we fake it
    #if hasattr(self,'_block'):
    if self._block.mode == 'query':
          try:
            value = self._block._queryValues[self]
          except KeyError:
            value = None
    elif self._block.mode == 'precommit':
        value = self._block._preCommitWorkingRecord.getField(self.field)

    else:
        value = self._block._resultSet.current.getField(self.field)
    # TODO: part of the removed block check from above
    #else:
    #  value = None

    if value == None:       value = ''

    # TODO: check to see if this is still needed
    #CheckBox does not work if comment following "if" block (ra3vat)
    if self.style == 'checkbox' and value == '':      value = 0

    return value

  #
  # triggerSetValue
  #
  # does proper typecasting when value is set via a trigger
  #
  def triggerSetValue(self, value):
    if self.style=='checkbox' and value =='':
      value = 0
    if self.typecast == "number" and value != None:
      value = float(value)
    self.setValue(value)

  def setValue(self, value):
    # TODO: This is now being done in the display handler
    #if self.style=='checkbox' and value =='':
    #  value = 0
    #
    # Number fields
    #if self.typecast == "number" and value != None:
    #  value = float(value)

    if self._block.mode == 'query':
      self._block._queryValues[self] = value

    elif self._block.mode == 'precommit':
      value = self._block._preCommitWorkingRecord.setField(self.field,value)

    else:
      self.processTrigger('Pre-Change')
      self._block._resultSet.current.setField(self.field,value)
      self.processTrigger('Post-Change')
    self._block._form.updateUIEntry(self)

  #
  # allowedValues
  #
  def allowedValues(self):
    if not self._allowedValues and hasattr(self,'_fk_datasource'):
      rs = self._fk_datasource.createResultSet()
      more = rs.firstRecord()
      self._allowedValues = {"":""}
      self._allowedValuesDescr = [""]
      self._allowedValuesReverse = {"":""}
      dropdownSeparator = gConfig('DropdownSeparator')
      while more:
        try:
          key = "%s" % rs.current.getField(self.fk_key)
          descriptions = []
          for i in self._fk_descr:
            descriptions.append("%s" % rs.current.getField(i))
          descr = string.join(descriptions,dropdownSeparator)

          self._allowedValues[key] = descr
          self._allowedValuesDescr.append(descr)
          self._allowedValuesReverse[descr] = key
        except AttributeError:
          # TODO: What's this for?
          self._allowedValues[rs.current.getField(self.fk_key)] = None
        more = rs.nextRecord()
      GDebug.printMesg (5,'Created for DropDown: %s' % self._allowedValues)

    return (self._allowedValues, self._allowedValuesDescr)

  def resetForeignKey(self):
    self._allowedValues = None
    self.allowedValues()


