#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2002 Free Software Foundation
#
# FILE:
# UIdriver.py
#
# DESCRIPTION:
# A curses based user interface driver for GNUe forms.
#
# NOTES:
#

from gnue.common.cursing import *
import curses
import atexit
from gnue.forms.GFForm import *
from gnue.common import events
from gnue.forms.uidrivers._base.UIdriver import *
from gnue.forms.uidrivers._base.UIWidget import *

##from UIapp import *

from gnue.forms import GFKeyMapper

import os
import sys
import string
from gnue.forms import VERSION
from gnue.common import GDebug, GLoginHandler
from gnue.common import GConfig
from gnue.common.TextUtils import lineWrap

_NOTEBOOK = None
_LOOPTRAP = 0
_EVENTPROCESSOR = None
_charWidth = 0
_IdToWxObj = {}
_IdToGFObj = {}
_IdToUIObj = {}


_SIO = None

def GetSIO():
  global _SIO
  if not _SIO:
    _SIO = sio()
    _SIO.BeginScreenAccess()
    atexit.register(_restore)
  return _SIO

#
# GFUserInterface
#
# The public interface to the User Interface
# All UIs must provide this class
#
class GFUserInterface(GFUserInterfaceBase):
  def __init__(self, eventController, disableSplash = None):
    self._WIDGETS = {'GFLabel'     : UILabel,
                     'GFBox'       : UIBox,
                     'GFPage'      : UIPage,
                     'GFForm'      : UIForm,
                     'GFEntry'     : UIEntry,
                     'GFButton'    : UIButton,
                     'GFScrollBar' : UIScrollBar,
                     }

    self._DIALOGS = {'about'       :UIAbout,
                     'messageBox'  :UIMessageBox,
                     }

    self._aboutDialog = UIAbout
    self._messageDialog = UIMessageBox

    self._disableSplash = disableSplash
    GFUserInterfaceBase.__init__(self,eventController)

    self._objectMapping = {}

    global _EVENTPROCESSOR
    _EVENTPROCESSOR = self.dispatchEvent

    self.forms = {}

    self.init()


  #
  # init
  #
  #
  def init(self):
    #
    # Create the main window
    #

    SIO = self._screen = GetSIO()
    GDebug.printMesg(1,'Initializing curses interface (TERM=%s)' % curses.termname())
    mw = self.mainWindow = ScreenContainer(SIO)
    SIO.Cls()

    #
    # SplashScreen
    #
    if not self._disableSplash:
      text = 'GNU Enterprise Forms\n' \
           + 'Version %s\n' % VERSION  \
           + '(c)2000-2002 Free Software Foundation\n' \
           + 'Protected by GNU General Public Licence 2.0\n'

#      MsgBoxOK(self.mainWindow, text).Show()


#    self.mainWindow.panel = MyDialog(self.mainWindow, 0, 0, self._screen.MAXROW, self._screen.MAXCOL )

    self.currentWidget = [self.mainWindow]


    menu = {'file':[ {'id'   :100,
                      'text' :'Commit     (F6)',
                      'tip'  : "Commit any pending changes",
                      'event':'requestCOMMIT'},
                     {'id'   :102,
                      'text' :'Rollback  (F11)',
                      'tip'  : "Rollback any pending changes",
                      'event':'requestROLLBACK'},
                     {'id': -1,'text': None, 'tip': None, 'event':None},
                     {'id'   :104,
                      'text' :'Print          ',
                      'tip'  : "Print the current screen",
                      'event':'requestPRINTOUT'},
                     {'id'   :101,
                      'text' :'Exit           ',
                      'tip'  : "Exit this application",
                      'event':"requestEXIT"},
                     ],
            'edit':[ {'id'   :401,
                      'text' :'Cut',
                      'tip'  :'Cut the selected text',
                      'event':'requestCUT'},
                     {'id'   :402,
                      'text' :'Copy',
                      'tip'  : "Copy the selected text",
                      'event':'requestCOPY'},
                     {'id'   :403,
                      'text' :'Paste',
                      'tip'  : "Paste the selected text",
                      'event':'requestPASTE'},
                     ],
            'data':[ {'id'   :201,
                      'text' :'Next Record',
                      'tip'  :"Go to the next record.",
                      'event':'requestNEXTRECORD'},
                     {'id'   :202,
                      'text' :'Previous Record',
                      'tip'  :"Go to the previous record.",
                      'event':'requestPREVRECORD'},
                     {'id'   :209,
                      'text' :'Jump to Record...',
                      'tip'  :"Go to a specific record.",
                      'event':'self.promptForRecordNumber'},
                     {'id': -1,'text': None, 'tip': None, 'event':None},
                     {'id'   :204,
                      'text' :'Delete Record          (F5)',
                      'tip'  :"Mark the current record as deleted.",
                      'event':'requestMARKFORDELETE'},
                     {'id'   :205,
                      'text' :'Next Block           (PgDn)',
                      'tip'  :"Go to the next logical block.",
                      'event':'requestNEXTBLOCK'},
                     {'id'   :206,
                      'text' :'Previous Block       (PgUp)',
                      'tip'  :"Go to the previous logical block",
                      'event':'requestPREVBLOCK'},
                     {'id': -1,'text': None, 'tip': None, 'event':None},
                     {'id'   :207,
                      'text' :'Enter Query            (F8)',
                      'tip'  :"Enter Query mode",
                      'event':'requestENTERQUERY'},
                      {'id'   :208,
                      'text' :'Execute Query        (F9)',
                      'tip'  :"Execute query",
                      'event':'requestEXECQUERY'},
                     ],
            'help':[ {'id'   :301,
                      'text' :'About...',
                      'tip'  :"Bring up a help screen regarding this application.",
                      'event':'requestAbout'},
                     ],
            }

    self.widgetWidth = 1  # The width of a 1 char widget (for things like buttons)
    self.widgetHeight = 1 # The pixel height of a 1 char widget (for things like buttons)
    self.textWidth = 1    # The pixel width of text inside a widget
    self.textHeight = 1   # The pixel height of text inside a widget


    self.menuBar = menuBar = MenuBar(self.mainWindow,0,0, self._screen.MAXCOL)
    self._menuIdMap = {}

    for _menu in ['file','edit','data','help']:
      newMenu=menuBar.NewMenu(_menu,_menu.capitalize())
      for menuitem in menu[_menu]:
        if menuitem['id'] > 0:
          newMenu.AddItem(menuitem['id'],menuitem['text'], menuitem['tip'])
          self._menuIdMap[menuitem['id']] = menuitem['event']
        else:
          newMenu.InsertSeperator()
      newMenu.SetMethod("CHOSEN",self._menuItemSelected)
      newMenu.Finalize()


    self.statusBar = StatusBar(self.mainWindow, "StatusBar",
      self._screen.MAXROW, 0,self._screen.MAXCOL, (-1,4,3,9,5))

    menuBar.SetStatusBar(self.statusBar,0)

    self.statusBar.SetText(0,'Welcome to GNUe Forms')
    self.statusBar.SetText(1,'OK')
    self.statusBar.SetText(2,'INS')
    self.statusBar.SetText(3,'0/0')
    self.statusBar.SetText(4,'1/1')

    self.mainWindow.AddControl(self.statusBar)
    self.mainWindow.Paint()

    return 1

  def activateForm(self, form):
    width = int(self._form.width)
    height = int(self._form.height)

##     formSize = wxSize(width*int(self.widgetWidth),
##                       int(height+self.menu_sb_space)*int(self.widgetHeight))

##     self.mainWindow.SetSize(formSize)
##     self.mainWindow.SetTitle(str(self._form.title))
    self.mainWindow.panel =  self._pageList[0]
    self.mainWindow.panel.Show()
##     self.mainWindow.panel.SetSize(formSize)
##     self._pageList[0].Show(1)

    # Only one page at a time can be visible
    self.visiblePage = self._pageList[0]

##     if _NOTEBOOK: # Adjust sizes
##       _NOTEBOOK.SetSize(formSize)

##       child = _NOTEBOOK.GetParent()
##       while child:
##         child.Fit()
##         child = child.GetParent()

##     self.mainWindow.CenterOnScreen()

    self.mainWindow.KeystrokeHook = keyboardEvtHandler(_EVENTPROCESSOR)._processEvent


  #############################################################################
  #
  # Private UIBase support functions
  #
  # Called in UIbase functions to perform UI interaction with this specific
  # widget set.
  #

  #
  # _exit
  #
  # Tells the application to close it's main window
  #
  def _exit(self):
##    self.mainWindow.EndScreenAccess()
    sys.exit()


  def _setStatusBar(self,*args):
    pass


  # Can be used by UI drivers to add more attributes to the creation event.
  def _updateCreateEvent(self, event):
    event.eventHandler = self.dispatchEvent
    event.ui = self
    event.formWidth = int(self._form.width)
    event.formHeight = int(self._form.height)
    event.screenHeight = self._screen.MAXROW - 3
    event.screenWidth = self._screen.MAXCOL
    return event


  #############################################################################
  #
  # Incoming Event Processors
  #
  # Processes the incoming events from other objects
  #

  #
  # mainLoop
  #
  # The primary loop of the user interface.  Called once the UI is
  # fully activated
  #
  def mainLoop(self):
    pass
    self.mainWindow.Run()

  def _setStatusBar(self,tip, statusValue, insertValue, currentRecord, maxRecord, currentPage, maxPage):
    if tip: self.statusBar.SetText(0,str(tip))

    if statusValue:
      self.statusBar.SetText(1,statusValue)

    if insertValue:
      self.statusBar.SetText(2,insertValue)

    if currentRecord and  maxRecord:
      self.statusBar.SetText(3,string.strip("%s/%s" % (currentRecord,maxRecord)))

    if currentPage and  maxPage:
      self.statusBar.SetText(4,string.strip("%s/%s" % (currentPage,maxPage)))


  #
  # gotoPage
  #
  # makes the requested page visible on the screen
  #
  def gotoPage(self, event):
    self.visiblePage.Hide()
    self.visiblePage = self._objectMapping[event.data]
    self.visiblePage.Show()

##     if not _NOTEBOOK:
##       self.visiblePage.Show(FALSE)
##       self.visiblePage = self._formToUI[event.data][0]
##       # TODO: move the calculatoin out of here
##       self.visiblePage.SetSize(wxSize(int(self._form.width)*int(self.widgetWidth),
##                                       int(int(self._form.height)+self.menu_sb_space)*int(self.widgetHeight)))
##     else:
##       self.visiblePage = self._formToUI[event.data][0]
##       for count in range(_NOTEBOOK.GetPageCount()):
##         if self.visiblePage.GetId() == _NOTEBOOK.GetPage(count).GetId():
##           _NOTEBOOK.SetSelection(count)
##           break
##     self.visiblePage.Show(1)

  #
  # formAlert
  #
  # Rings a bell and alters the statusbar to display
  # a line of text
  #
  def formAlert(self, event):
    curses.beep()
    curses.flash()


  # Called whenever forms goes into a "wait" state in which user cannot
  # interact with interface (e.g., while waiting for a query or a commit)
  def beginWait (self, event):
    pass


  # Called whenever forms leaves a "wait" state
  def endWait (self, event):
    pass

  def activateMenu(self, event):
    GDebug.printMesg(0,'MENU ACTIVATED')
    self.menuBar.Run()


  def _menuItemSelected(self, menu, menuid, title):
    action = self._menuIdMap[menuid]
    GDebug.printMesg(0,'***Menu item %s selected %s' % (menuid, action))
    self.dispatchEvent(events.Event(action))
    self.dispatchEvent(events.Event('gotoENTRY', self._form._currentEntry))

  #############################################################################
  #
  # Internal Event Processors
  #
  # Processes the events from the widget set
  #

  #
  # menuExitEvent
  #
  # Catches the Exit event from the menu and make sure it closes in such a way to
  # trigger the closeTrap method
  #
  def menuExitEvent(self, event):
    self.mainWindow.Close()

  #
  # promptForRecordNumber
  #
  # Presents a dialog to user asking for the record number to jump to
  #
  def promptForRecordNumber(self, event= None):
     pass


  #
  # closeTrap
  #
  # intercepts the applications closure and generates an event to the form requesting
  # closure.  Allows the form to control closure.  If the form approves it'll send
  # back an event closing the application
  #`
  def closeTrap(self,event):
    pass
##     object = _eventObjTowxWindow(event)
##     if event.CanVeto():
##       self.dispatchEvent(events.Event('requestEXIT'))
##     else:
##       object.Destroy()



#############################################################################
#############################################################################
#
# UI Objects
#
# A set of standard UI objects that allow us to move more code into UIbase
# These map generic functionality to the specific UI widget set in use by
# this driver.
#

#############################################################################
#############################################################################

#
# UIHelper
#
# Provides a UI widget set of std functions to reduce the number of functions
# require in each UIclass listed below
class UIHelper:
  def showModal(self):
    pass # self.ShowModal()

  def show(self, flag=1):
    pass # self.Show(flag)

  def destroy(self):
    pass # self.Destroy()

  def indexedFocus(self, index):
    widget = self.widgets[index]
    widget.PARENT.SetFocus(widget,run=0)

  def setValue(self, value, index=0, enabled=1):
    self.widgets[index].SetValue(value)
##     self.widgets[index].SetEvtHandlerEnabled(FALSE)
##     self.widgets[index].SetValue(value)
##     self.widgets[index].SetEvtHandlerEnabled(1)

  def setCursorPosition(self, position, index=0):
    try:
      self.widgets[index].SetCursor(position)
    except AttributeError:
      pass

  def setSelectedArea(self, selection1, selection2, index=0):
    pass
##     try:
##       self.widgets[index].SetSelection(selection1, selection2)
##     except AttributeError:
##       pass  # For label-style entries

  def _addToCrossRef(self, widget,gfobject, uiobject):
    pass
##     id = widget.GetId()
##     _IdToWxObj[id]=widget
##     _IdToGFObj[id]=gfobject
##     _IdToUIObj[id]=uiobject

  def createWidget(self, event, spacer):

    newWidget = self._createWidget(event, spacer)
    self._addToCrossRef(newWidget,event.object,self)
    return newWidget

#
# UILabel
#
# Widget set specific function that creates a single instance of a label
#
class UILabel(UIHelper, UIWidget):
  alignmentStyle = { 'left'  : 0,
                     'center': 0,
                     'right' : 0,
                     }

  def _createWidget(self, event, spacer):
    object = event.object
    style = self.alignmentStyle[object.alignment]
    newWidget = Label(event.container,str(self),object.y+spacer,object.x,str(object.text))
    event.container.AddControl(newWidget)
    return newWidget

#
# UIBox
#
# Widget set specific function that creates a single instance of a border box
#
class UIBox(UIHelper, UIWidget):
  def _createWidget(self, event, spacer):
    object = event.object
    newWidget = GroupBox(event.container,'name',object.y,object.x,
                      object.width, object.height,str(object.label))
    event.container.AddControl(newWidget)
    return newWidget

#
# UIScrollBar
#
# Widget set specific function that creates a single instance of a scroll bar used
# to navigate multiple records
#
class UIScrollBar(UIHelper, UIWidget):
  def _createWidget(self, event, spacer):
    object = event.object

    if object.width > 1:
      newWidget = ScrollBar(event.container,'name',object.y,object.x,object.width)
    else:
      newWidget = VScrollBar(event.container,'name',object.y,object.x,object.height)
    event.container.AddControl(newWidget)
    newWidget.Init(10)  ## TODO: This should be the number of rows
    return newWidget

#
# UIButton
#
# Widget set specific function that creates a single instance of a button
#
class UIButton(UIHelper, UIWidget):
  def _createWidget(self, event, spacer):
    object = event.object
    newWidget = Button(event.container,'name',object.y+spacer,object.x,object.width,str(object.label))
    event.container.AddControl(newWidget)
    self.__trigger = object.trigger
    newWidget.SetMethod("CLICK", self.buttonHandler)
    return newWidget

  def buttonHandler(self, a1, a2, a3):
    GDebug.printMesg(1,"Click!")
##     if event.GetEventType() == wxEVT_COMMAND_BUTTON_CLICKED:
##       action = None
##       object = _eventObjTowxWindow(event)
##       id = object.GetId()
##       gfObject     = _IdToGFObj[id]

    _EVENTPROCESSOR(events.Event('fireTRIGGER',self.__trigger))

##     if action:
##       _EVENTPROCESSOR(action)

#
# UIPage
#
# Widget set specific function that creates a single instance of a page
#
# Note: The event trap for changing notebook pages is attached to the UIForm
#
class UIPage(UIHelper, UIWidget):
  def _createWidget(self, event, spacer):
    object = event.object

    width = event.formWidth + 2
    height = event.formHeight + 2
    x = int(event.screenWidth/2 - width/2)
    y = int(event.screenHeight/2 - height/2)

    # For the menu bar
    if y < 2:
      y = 2
    # For sanity
    if x < 0:
      x = 0
    if y + height > event.screenHeight - 3:
      height = event.screenHeight - 3
    if x + width > event.screenWidth- 2:
      width = event.screenWidth


    newWidget = Frame(event.container, y, x,  y+height-1, x+width-1 , object._parent.title)
    ###newWidget.DeActivate(0)
    event.interface._pageList.append(newWidget)
    event.interface._objectMapping[object] = newWidget
    event.interface._objectMapping[newWidget] = object

    ####
    newWidget.KeystrokeHook = keyboardEvtHandler(_EVENTPROCESSOR)._processEvent

    return newWidget

#
# UIEntry
#
# Widget set specific function that creates a single instance of a data entry widget
#
class UIEntry(UIHelper, UIWidget):
  def _createWidget(self, event, spacer):
    object = event.object

    style = object.style
    if style == 'dropdown':
      if event.initialize:
        choices = object.allowedValues()[1]
      else:
        choices = [""]
      newWidget = ComboBox(event.container, 'name',object.y+spacer,object.x, object.width)
      for index in range(len(choices)):
        newWidget.AddItem(index,choices[index])
    elif style == 'label':
      newWidget = TextBox(event.container, 'name', object.y+spacer, object.x, object.width)
    elif style == 'checkbox':
      newWidget = CheckBox(event.container, 'name',object.y+spacer,object.x,'',0)
    elif object.height > 1: # Normal text box/Multiline
      newWidget = TextArea(event.container, 'name', object.y+spacer, object.x, object.width, object.height)
    else: # Normal text box
      newWidget = TextBox(event.container, 'name', object.y+spacer, object.x, object.width)

    event.container.AddControl(newWidget)

    return newWidget

#
# UIForm
#
# Widget set specific function that creates a single instance of a Form widget
#
class UIForm(UIHelper, UIWidget):
  tabStyles = {'top':0,
               'left':1,
               'right':2,
               'bottom':3,
               }

  def _createWidget(self, event, spacer):
    object = event.object
##    newWidget = Dialog(event.container, 0, 0, event.interface._screen.MAXROW, event.interface._screen.MAXCOL )

##    return newWidget


#####################################################################
##
## Required Dialog box support
##
#####################################################################

#
# UIAbout
#
# A dialog box providing about info
#
class UIAbout( UIHelper):
  def __init__(self, interface, programVersion, formName, formVersion, author, description):
    pass

# UImessageBox
#
# A dialog box providing basic message info
#
class UIMessageBox(UIHelper):
  def __init__(self, interface, message, caption):
    message = lineWrap(message,60)
    pass

#####################################################################
##
## Keymapper Support
##
#####################################################################
from gnue.forms.GFKeyMapper import vk

# Translate from curses keystrokes to our virtual keystrokes
cursesKeyTranslations = {
  vk.F1        : curses.KEY_F1,        vk.F2       : curses.KEY_F2,
  vk.F3        : curses.KEY_F3,        vk.F4       : curses.KEY_F4,
  vk.F5        : curses.KEY_F5,        vk.F6       : curses.KEY_F6,
  vk.F7        : curses.KEY_F7,        vk.F8       : curses.KEY_F8,
  vk.F9        : curses.KEY_F9,        vk.F10      : curses.KEY_F10,
  vk.F11       : curses.KEY_F11,       vk.F12      : curses.KEY_F12,
  vk.INSERT    : curses.KEY_IC,        vk.DELETE   : curses.KEY_DC,
  vk.HOME      : curses.KEY_HOME,      vk.END      : curses.KEY_END,
  vk.PAGEUP    : curses.KEY_PPAGE,     vk.PAGEDOWN : curses.KEY_NPAGE,
  vk.UP        : curses.KEY_UP,        vk.DOWN     : curses.KEY_DOWN,
  vk.LEFT      : curses.KEY_LEFT,      vk.RIGHT    : curses.KEY_RIGHT,
  vk.TAB       : 9,                    vk.ENTER    : 10,
  vk.BACKSPACE : curses.KEY_BACKSPACE }


GFKeyMapper.KeyMapper.setUIKeyMap(cursesKeyTranslations)

#####################################################################
##
## Basic Event Processing
##
#####################################################################

def _setDefaultEventHandlers(newWidget, initialize):
  global _EVENTPROCESSOR
  if initialize:
    newWidget.SetMethod("CLICK", mouseEvtHandler(_EVENTPROCESSOR))


class uiBaseEvtHandler:
  def __init__(self,eventList,eventProcessor):
    pass
##     wxEvtHandler.__init__(self)
##     self._eventProcessor = eventProcessor
##     for eventType in eventList:
##       self.Connect(-1, -1, eventType,self._processEvent)

##   def _processEvent(self,a1, a2, a3):
##     GDebug.printMesg(0,"uiBaseEvtHandler _processEvent was called?!?")

class mouseEvtHandler(uiBaseEvtHandler):
  def __init__(self,eventProcessor):
    pass
##     eventList = [wxEVT_LEFT_DOWN]
##     uiBaseEvtHandler.__init__(self, eventList, eventProcessor)

##   def _processEvent(self,event):
##     global _charWidth
##     object = _eventObjTowxWindow(event)

##     # compute the location of the character in the widget
##     x,y = event.GetPosition()
##     cursorPosition = 1 + x/_charWidth

##     # Move to proper object
##     id = object.GetId()
##     gfObject     = _IdToGFObj[id]
##     screenWidget = _IdToWxObj[id]
##     count        = _IdToUIObj[id].widgets.index(screenWidget)
##     self._eventProcessor(events.Event('requestFOCUS',gfObject))
##     self._eventProcessor(events.Event('requestJUMPRECORD',count - gfObject._visibleIndex))
##     self._eventProcessor(events.Event('requestCURSORMOVE',position=cursorPosition))

##     event.Skip()

class keyboardEvtHandler(uiBaseEvtHandler):
  def __init__(self, eventProcessor):
    self._eventProcessor = eventProcessor
    uiBaseEvtHandler.__init__(self, [], eventProcessor)

  def _processEvent(self, keystroke):
    ####return keystroke
    action = None

    ##object = _eventObjTowxWindow(event)


    GDebug.printMesg(0,"key====== %s" % keystroke)
    command = GFKeyMapper.KeyMapper.getEvent(
        keystroke,
        0, #event.ShiftDown(),
        0, #event.ControlDown(),
        0) #event.AltDown())

    GDebug.printMesg(0,"========= %s" % command)

      # TODO : Broken for the moment, this should probably be form driven
      #if command == 'JUMPRECORD':
      #  self.promptForRecordNumber()

    if command:
      action = events.Event('request%s' % command)

    else:
      try:
        GDebug.printMesg(1,'Keystroke=(%s,%s)' % (keystroke, chr(keystroke)))
        if 32 <= keystroke <= 255 and chr(keystroke) in string.printable:
          GDebug.printMesg(1,'Keystroke is printable')
          action = events.Event('requestKEYPRESS', chr(keystroke),
                         text=chr(keystroke),
                         code=keystroke)
      except ValueError:
        pass

    if action:
      self._eventProcessor(action)

    return None


#####################################################################
##
## Login Support
##
#####################################################################

#
# UILoginHandler
#
class UILoginHandler(GLoginHandler.LoginHandler):
##   def __init__(self):
##     self._wxapp = getWxApp()
##     self.dlg = None

   def getLogin(self, loginData, errorText=None):
     if 1: # TODO: This is left here to prevent me from having to unindent this code
           # TODO: Since the UI stuff is currently being gutted I'm not wasting the time
           # TODO: to do the right thing

       # Get the screen
       root = GetSIO()
       screen = ScreenContainer(root)

       if len(loginData[1]):
         loginMesg = 'Login required for "%s"' % (loginData[1])
       else:
         loginMesg = 'Login required for %s' % (loginData[0])

       gnueMesg = _("GNU Enteprise Forms")

       dlgWidth = max(len(loginMesg), len(gnueMesg)) + 4
       dlgHeight = 5

       labelList = []
       textctrlList = []
       labelMsgList = []

       fieldLabelWidth = 0

       for prompt in loginData[2]:
         labelMsgList.append('%s:' % prompt[1])
         fieldLabelWidth = max(fieldLabelWidth, len ('%s:' % prompt[1]))
         dlgWidth = max(dlgWidth, fieldLabelWidth + 23)
         dlgHeight += 1

       lbMesg = _('Login')
       cbMesg = _('Cancel')
       dlgWidth = max(dlgWidth, len(lbMesg) + len(cbMesg) + 11) + 2
       dlgHeight += 1

       #
       # Create widgets
       #

       w = dlgWidth + 2
       h = dlgHeight + 2
       x = int(root.MAXCOL/2 - w/2 + .5)
       y = int(root.MAXROW/2 - h/2 + .5)

       dlg = Dialog(screen, y, x,  y+h-1, x+w-1)

       gnueLabel = dlg.AddDialogControl(Label(dlg, "lbl1", 1,
                               int((dlgWidth - len(gnueMesg))/2),
                               gnueMesg))
       loginLabel = dlg.AddDialogControl(Label(dlg, "lbl1", 2,
                               int((dlgWidth - len(loginMesg))/2),
                               loginMesg))

       firstY = 4
       lastY = firstY
       xPos = int(dlgWidth/2 - fieldLabelWidth/2) - 11

       for prompt in loginData[2]:
         s = dlg.AddDialogControl(Label(dlg,"lbl2",lastY, xPos,'%s:' % prompt[1]))
         labelList.append(s)
         t = dlg.AddDialogControl(TextBox(dlg, "inp", lastY, xPos + fieldLabelWidth + 2, 20, Password=prompt[2]))
         textctrlList.append(t)
         lastY += 1

       loginButton = dlg.AddDialogControl(Button(dlg,"LoginButton",dlgHeight - 1,dlgWidth - len(cbMesg) - len(lbMesg) - 10,len(lbMesg)+4,lbMesg))
       loginButton.EXITFORM = 1
       cancelButton = dlg.AddDialogControl(Button(dlg,"CancelButton",dlgHeight - 1,dlgWidth - len(cbMesg) - 5,len(cbMesg)+4,cbMesg))
       cancelButton.EXITFORM = 1

       t = dlg.RunDialog()

       if dlg.Controls[t] == cancelButton:
         raise GLoginHandler.UserCanceledLogin

       rv = {}
       for i in range(0, len(loginData[2])):
         rv[loginData[2][i][0]] = textctrlList[i].TEXT

       return rv


def _restore(*args, **parms):
  # Set everything back to normal
  try:
#    _SIO.keypad(0)
    curses.echo()
    curses.nocbreak()
    curses.endwin()          # Terminate curses
  except:
    pass



